#include "ReciprocalToRealProjection.h"

namespace pinkIndexer
{

    ReciprocalToRealProjection::ReciprocalToRealProjection(const ExperimentSettings& experimentSettings)
        : experimentSettings(experimentSettings)
    {
    }
} // namespace pinkIndexer
