/*
 * tests.h
 *
 *  Created on: 04.05.2017
 *      Author: Yaro
 */

#ifndef TESTS_H_
#define TESTS_H_

namespace pinkIndexer
{
    void testPinkIndexer();
    void testPatternPrediction();
    void testRefinement();
    void testRefinementGetDefect();
    void testSinogramComplete();
    void testSinogram2();
    void testSinogram();
    void testReflectionsInRangeFinder();
    void testBackprojection();
    void testSimpleProjection();
} // namespace pinkIndexer

#endif /* TESTS_H_ */
